from __future__ import print_function

from IPython.display import IFrame, display
from ipywidgets import Box, Output
from ginga.web.pgw import ipg
from astropy.io.fits import HDUList

__all__ = ['ImageViewer']


class ImageViewer(Box):
    """
    Basic image viewer Jupyter widget which wraps a ginga widget.

    Set the content of the image viewer via the ``image`` attribute.

    Parameters
    ----------

    image : str, FITS HDU, or numpy array
        Content to display in the widget

    """
    _server = ipg.make_server(host='localhost',
                                       port=9987,
                                       use_opencv=False)
    _server_started = False
    _number_views = 0

    def __init__(self, image=None, width=600, height=600,
                 show_position=False, **kwd):
        super(ImageViewer, self).__init__(**kwd)

        self._image_display = Output(value="I will have an image soon")


        self.children = [self._image_display]
        if not ImageViewer._server_started:
            ImageViewer._server.start(no_ioloop=True)
            ImageViewer._server_started = True
        #self._server = None
        self._viewer = None
        self._width = width
        self._height = height
        self._show_position = show_position
        if self._show_position:
            self._height += 30
        # This call sets both server and viewer
        self._ginga_viewer()

        width, height = self._viewer.get_window_size()

        with self._image_display:
            display(IFrame(self._viewer.url, width, height))

        self.image = image

    def _ginga_viewer(self):
        # Set this to True if you have a non-buggy python OpenCv
        # bindings--it greatly speeds up some operations
        use_opencv = False

        if self._show_position:
            viewer_class = ipg.EnhancedCanvasView
        else:
            viewer_class = ipg.BasicCanvasView

        ImageViewer._number_views += 1
        self._viewer = self._server.get_viewer(str(ImageViewer._number_views),
                                               width=self._width,
                                               height=self._height,
                                               viewer_class=viewer_class)
        self._viewer.set_zoom_algorithm('rate')
        self._viewer.set_zoomrate(1.01)

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, value):
        self._image = value
        if value is not None:
            if isinstance(value, str):
                # Hope this is the name of a FITS file!
                self._viewer.load_fits(value)
            elif isinstance(value, HDUList):
                self._viewer.load_hdu(value)
            else:
                # Sure hope this is really numpy data
                self._viewer.load_data(value)



from astropy.stats import sigma_clipped_stats
from photutils import DAOStarFinder
from photutils import fit_2dgaussian
from astropy.nddata.utils import Cutout2D
from astropy.stats import gaussian_sigma_to_fwhm

__all__ = ['source_detection']


def source_detection(ccd, fwhm=3.0, sigma=3.0, iters=5,
                     threshold=5.0, find_fwhm=True):
    """
    Returns an astropy table containing the position of sources
    within the image.

    Parameters
    ----------------

    ccd : numpy.ndarray
        The CCD Image array.

    fwhm : float, optional
        Full-width half-max of stars in the image.

    sigma : float, optional.
        The number of standard deviations to use as the lower and
        upper clipping limit.

    iters : int, optional
        The number of iterations to perform sigma clipping

    threshold : float, optional.
        The absolute image value above which to select sources.

    find_fwhm : bool, optional
        If ``True``, estimate the FWHM of each source by fitting a 2D Gaussian
        to it.

    Returns
    -----------

    sources
        an astropy table of the positions of sources in the image.
        If `find_fwhm` is ``True``, includes a column called ``FWHM``.
    """
    data = ccd
    mean, median, std = sigma_clipped_stats(data, sigma=sigma, iters=iters)
    daofind = DAOStarFinder(fwhm=fwhm, threshold=threshold * std)
    sources = daofind(data - median)
    if find_fwhm:
        fwhm_fit = []
        for source in sources:
            x = source['xcentroid']
            y = source['ycentroid']
            cutout = Cutout2D(data, (x, y), 5 * fwhm)
            fit = fit_2dgaussian(cutout.data)
            fwhm_fit.append(gaussian_sigma_to_fwhm * (fit.x_stddev + fit.y_stddev) / 2)
        sources['FWHM'] = fwhm_fit
    return sources




# coding: utf-8

# In[1]:

from __future__ import print_function, division

import numpy as np

from astropy.visualization import scale_image
from astropy.nddata import block_replicate
from astropy.table import Table, Column
import astropy.units as u

import astropy.coordinates as apycoord

from catalog_search import catalog_search
from filter_transform_ivezic_2007 import filter_transform

from ccdproc import CCDData



import matplotlib.pyplot as plt
get_ipython().magic(u'matplotlib inline')


# # Read image

# In[2]:

ccd = CCDData.read('kelt-1-b-054R.fit', unit='adu')


# In[3]:

# This function really should be somewhere else eventually.
def scale_and_downsample(data, downsample=4,
                         min_percent=20,
                         max_percent=99.5):

    scaled_data = scale_image(data,
                              min_percent=min_percent,
                              max_percent=max_percent)

    if downsample > 1:
        scaled_data = block_reduce(scaled_data,
                                   block_size=(downsample, downsample))
    return scaled_data

def uniformize_source_names(aij_tbl):
    import re
    
    data_col = re.compile(r'Source-Sky_[TC](\d+)')
    sources = []
    for c in aij_tbl.colnames:
        match = re.search(data_col, c)
        if match:
            sources.append(len(sources) + 1)
            source_number =  match.groups()[0]
            try:
                aij_tbl.rename_column(c, 'Source-Sky_C' + source_number)
            except KeyError:
                # Column already exists s there was no need to change the name
                pass
            try:
                aij_tbl.rename_column('Source_Error_T' + source_number,
                                      'Source_Error_C' + source_number)
            except KeyError:
                pass
            try:
                aij_tbl.rename_column('Peak_T' + source_number,
                                      'Peak_C' + source_number)
            except KeyError:
                pass
            try:
                aij_tbl.rename_column('RA_T' + source_number,
                                      'RA_C' + source_number)
            except KeyError:
                pass
            try:
                aij_tbl.rename_column('DEC_T' + source_number,
                                      'DEC_C' + source_number)
            except KeyError:
                pass
            try:
                aij_tbl.rename_column('X(FITS)_T' + source_number,
                                      'X(FITS)_C' + source_number)
            except KeyError:
                pass
            try:
                aij_tbl.rename_column('Y(FITS)_T' + source_number,
                                      'Y(FITS)_C' + source_number)
            except KeyError:
                pass
            
    return sources

def source_column(source_number):
    col_name = 'Source-Sky_C' + str(source_number)
    return col_name

def source_ra(source_number):
    col_name = 'RA_C'+str(source_number)
    return col_name

def source_dec(source_number):
    col_name = 'DEC_C'+str(source_number)
    return col_name

def source_x_pix(source_number):
    col_name = 'X(FITS)_C'+str(source)
    return col_name
    
def source_y_pix(source_number):
    col_name = 'Y(FITS)_C'+str(source)
    return col_name

def source_error(source_number):
    col_name = 'Source_Error_C' + str(source_number)
    return col_name


# In[4]:

apass, apass_x, apass_y = catalog_search(ccd.wcs, ccd.shape, 'II/336/apass9', 'RAJ2000', 'DEJ2000')
apass_bright = (apass['e_r_mag'] < 0.05) & (apass['u_e_r_mag'] == 0) & (apass['e_B-V'] < 0.1)
apass_in_bright, in_apass_x, in_apass_y = apass[apass_bright], apass_x[apass_bright], apass_y[apass_bright]

vsx, vsx_x, vsx_y = catalog_search(ccd.wcs, ccd.shape, 'B/vsx/vsx', 'RAJ2000', 'DEJ2000')
vsx_names = vsx['Name']
print(ccd.shape)
print(type(ccd))


# In[5]:

disp = scale_and_downsample(ccd.data, downsample=1)


# In[6]:

import mpld3
mpld3.enable_notebook()
#mpld3.disable_notebook()


# In[7]:

plt.figure(figsize=(12, 7))
plt.imshow(disp, cmap='gray', origin='lower')
plt.scatter(vsx_x, vsx_y, c='none', s=100, edgecolor='cyan')
plt.title('Blue: VSX, Yellow: APASS', fontsize=20)

for x, y, m in zip(vsx_x, vsx_y, vsx_names):
    plt.text(x, y, str(m), fontsize=18, color='cyan')
    
plt.scatter(in_apass_x, in_apass_y, c='none', s=50, edgecolor='yellow', alpha=0.5, marker='o')
#for x, y, c in zip(apass_x, apass_y, apass_in):
#    plt.text(x, y, c.to_string(), fontsize=12, color='yellow')

plt.xlim(0, ccd.shape[1])
plt.ylim(0, ccd.shape[0])


# In[8]:

apass_mags = apass_in_bright['r_mag']
apass_ra = apass_in_bright['RAJ2000']
apass_dec = apass_in_bright['DEJ2000']
apass_color = apass_in_bright['B-V']

aij_raw = Table.read('Measurements_2013-09-13_R_clipping.csv')
image_index = [el for el in range(len(aij_raw['dumbname']))]
sources = uniformize_source_names(aij_raw)
source_index = [el for el in range(len(sources))]
aij_ra = []
aij_dec = []
aij_mags = []
for source in sources:
    aij_ra.append(aij_raw[source_ra(source)][len(aij_raw['dumbname'])//2])
    aij_dec.append(aij_raw[source_dec(source)][len(aij_raw['dumbname'])//2])
    aij_mags.append(-2.5*np.log10(aij_raw[source_column(source)]))
aij_mags = np.array(aij_mags)
aij_coordinates = apycoord.SkyCoord(aij_ra, aij_dec, unit=(u.hour, u.deg))
apass_coordinates = apycoord.SkyCoord(apass_ra, apass_dec, unit='deg')


# In[9]:

match = apycoord.match_coordinates_sky(aij_coordinates, apass_coordinates)


# In[ ]:




# #APASS Filter Corrections
# ##Transform the APASS r magnitudes into R magnitudes using APASS r and i magnitudes
# 
# The equation used is R-feder - r-apass = A*c**3 + B*c**2 + C*c + D
# 
# Where...

# In[10]:

A =-0.0107
B = 0.0050
C =-0.2689
D =-0.1540


# And the value c is dependant on each object and is equal to the i magnitude subtracted from the r magnitude. 
# 
# c = r-i
# 
# The initial equation can be solved for BVRIfeder by simply adding griapass on both sides.
# 
# R-feder = Ac**3 + Bc*2 + Cc + D + r-apass

# In[11]:

apass_r_mags = apass_in_bright['r_mag']
apass_i_mags = apass_in_bright['i_mag']
c = apass_r_mags - apass_i_mags
apass_R_mags = (A*(c**3))+(B*(c**2))+(C*c)+D+apass_r_mags


# In[ ]:




# In[12]:

transformed = filter_transform(apass_in_bright, 'R', r='r_mag', i='i_mag')
transformed.description


# In[13]:

print(transformed)


# In[14]:

get_ipython().magic(u'matplotlib inline')
corrections = []
fit_error = []
for image in image_index:
    BminusV = []
    Rminusr = []
    for index, el in enumerate(match[0]):
        if aij_coordinates[index].separation(apass_coordinates[el]).arcsec < 0.5:
            if aij_mags[index][image] < 100:
                BminusV.append(apass_color[el])
                Rminusr.append(apass_R_mags[el]-aij_mags[index][image])
    slope_intercept, cov = np.polyfit(BminusV, Rminusr, 1, cov=True)
    corrections.append(slope_intercept)
    """The Error in the slope and the intercept are the diagonals of the matrix cov. 
    The first diagonal ([0,0]) is the error in the slope and the second diagonal
    is the error in the intercept. See the Documentation for polyfit for more info"""
    fit_error.append([np.sqrt(cov[0,0]), np.sqrt(cov[1,1])])


# In[16]:


good_pics = image_index
slope_error = []
intercept_error = []
for el in fit_error:
    slope_error.append(el[0])
    intercept_error.append(el[1])


# In[17]:

kelt_1_data = []
keltBminusV = 0.6    ##########################MUST HAVE B-V FOR ALL STARS IN FIELD!!!!#####################
corrected_curves = np.zeros((len(sources), len(good_pics)))
corrected_curves_er = np.zeros((len(sources), len(good_pics)))
all_SNR = np.zeros((len(good_pics),len(sources)))
for obj in source_index:
    tmp_data = []
    tmp_data_er = []
    plt.figure()
    for image in good_pics:
        ###Date
        JD = aij_raw['J.D.-2400000'][image]
        ###Magnitude
        color_term = keltBminusV*corrections[image][0]
        intercept = corrections[image][1]
        old_mag = aij_mags[obj][image]
        new_mag = old_mag+intercept+color_term
        tmp_data.append(new_mag)
        ###Error
        ##Error in color term
        #Error in slope
        frac_unc_slope = slope_error[image]/corrections[image][0]
        #Error in B-V
        er_BV = 0.005   #Estimation
        frac_unc_BV = er_BV/keltBminusV
        
        color_err = np.sqrt(frac_unc_slope**2+frac_unc_BV**2)*color_term
        ##Error in intercept term
        er_intercept = intercept_error[image]
        ##Error in old mag
        SNR = aij_raw[source_column(obj+1)][image]/aij_raw[source_error(obj+1)][image]
        er_old_mag = 1/SNR
        all_SNR[image,obj] = 1.0*SNR
        
        er = np.sqrt(er_old_mag**2+er_intercept**2+color_err**2)
        tmp_data_er.append(er)
        ###Plot
        plt.errorbar(JD, new_mag, yerr=er, color='grey')
        plt.errorbar(JD, new_mag, color='red', xerr=0.0005)
    corrected_curves[obj] = np.array(tmp_data)
    corrected_curves_er[obj] = tmp_data_er
    #plt.title('Source_C'+str(obj+1))
    #plt.show()


# ###Check what the slope intercepts and slope values are over the course of the night

# In[34]:

#Check the slope intercept over the night
for index, slopeinfo in enumerate(corrections):
    plt.scatter(index, slopeinfo[1])
plt.title('intercept')
plt.show()

#Check the slope values over the night
for index, slopeinfo in enumerate(corrections):
    plt.scatter(index, slopeinfo[0])
plt.title('slope')
plt.show()


# #Finding Comparison Stars
# ##An automated way of finding the best comparison stars from the apass calibrated data
# 
# For each source, the standard deviation is calculated. This deviation is the mulitplied by the sources average magnitude. The sources with the lowest values have an optimal combination of being fairly stable and fairly bright.

# #Comparison Set Calculations
# ## With Error

# In[76]:

def calc_comp_weights(images, comparison_sources, SNR_data):
    #Calculate the initial weights for each source for each image by deviding the sources SNR^2 by the norm const for that image
    """norm_consts = np.zeros(len(images))
    for image in images:
        #set the sum equal to zero
        SNR2sum = 0
        for source in comparison_sources:
            #add the SNR^2 to the sum
            SNR2sum += SNR_data[image-1,source-1]**2
        #set the norm constant for that image equal to one over the SNR^2 sum
        norm_consts[image-1] = 1/SNR2sum
        
    weights = np.zeros((len(images),len(comparison_sources)))
    for image, norm in zip(images, norm_consts):
        for source in comparison_sources:
            weights[image-1,source-1] = norm*(SNR_data[image-1,source-1]**2)"""
    
    comparison_sources = np.array(comparison_sources)
    norm_const = 1/np.sum((SNR_data[0,:]**2))
    
    weights = np.zeros(len(comparison_sources))
    for source in comparison_sources:
        weights[source-1] = norm_const*(SNR_data[0,source-1]**2)
    return weights

def calc_comp_star(images, comparison_sources, mag_data, weights):
    vert_comp_star = np.zeros(len(images))
    for image in images:
        comp_sum = 0
        for source in comparison_sources:
            if source_list[source-1]:
                comp_sum += weights[source-1]*mag_data[source-1][image-1]
        vert_comp_star[image-1] = comp_sum

    vert_comp_star = vert_comp_star/np.sum(weights, axis = 1)
    
    return vert_comp_star


# In[77]:

"""
Calculate constant that normalizes the set of weights for each of the images (One normalization constant for each image)
This constant will be used to calculate the weights for each source in each image. This weight is calculated by deviding the
sources SNR^2 by the normalization constant for that image.
"""
source_list = [True for el in sources]
source_list[1] = False

comparison_sources = sources#[4,5,8,9,10]

weights = calc_comp_weights(good_pics, comparison_sources, all_SNR)

vert_comp_star = calc_comp_star(good_pics, comparison_sources, corrected_curves, weights)

"""
7/13/2016

The issue you had last night is that the vert_comp_star is varying by many magnitudes, leading you to beleive there
is an error in the way you are calculating the comparison star. Note that the weights ARE all normalized for each
image so that part at least appears to be correct.

Solved
"""

#now we calculate the differential magnitudes
first_differential = np.zeros((len(good_pics),len(sources)))
for source in sources:
    first_differential[:,source-1] = corrected_curves[source-1] - vert_comp_star

"""
######     ##         #####    ##   ##      ##    ##    ######       ##    ##    ########    ######     ########
#######    ##     #########    ##  ##       ##    ##    #######      ##    ##    ##          #######    ##
##   ##    ##    ###           ## ##        ##    ##    ##   ##      ##    ##    ##          ##   ##    ##
######     ##    ##            ####         ##    ##    ######       ########    #######     ######     #######
##         ##    ###           ####         ###  ###    ##           ##    ##    ##          ##   ##    ##
##         ##     #########    ## ###        ######     ##           ##    ##    ##          ##   ##    ##
##         ##         #####    ##   ##        ####      ##           ##    ##    ########    ##   ##    ########

7/13/2016

You think you got the first differential photometry right, now it's time to make the iterative process.
"""

plt.plot(vert_comp_star, 'o')


# In[ ]:




# In[19]:

"""
Calculate constant that normalizes the set of weights for each of the images (One normalization constant for each image)
This constant will be used to calculate the weights for each source in each image. This weight is calculated by deviding the
sources SNR^2 by the normalization constant for that image.
"""
norm_consts = np.zeros(len(good_pics))
for image in good_pics:
    SNR2sum = 0
    for source in source_index: 
        SNR2sum += all_SNR[image-1,source]**2
    norm_consts[image] = 1/SNR2sum

#Calculate the initial weights for each source for each image by deviding the sources SNR^2 by the norm const for that image
weights = np.zeros((len(good_pics),len(sources)))
for image, norm in enumerate(norm_consts):
    for source in source_index:
        weights[image,source] = norm*(all_SNR[image,source]**2)

"""
Now the weights should all be calculated... Not too positive on if we should include all sources in this or not but I'm
hoping that the value for the norm constant will eventually make the sources with largest SNR sorta have a small enough
weight that it doesn't matter but we may have to have a larger loop outside of this that takes out certain stars from
the comparison set that are screwing everything up... See figure 1 in Broeg et al. for sorta more clarification (syntax
in that figure is weird).

The weights can now be used to calculate the vertual comparison star
"""

######Make sure that when this is all over every array like corrected_curves is a flat array (if possible).

vert_comp_star = np.zeros(range(goodPics))
for image in good_pics:
    comp_sum = 0
    for source in source_index:
        comp_sum += weights[image, source]*corrected_curves[source][image]
    vert_comp_star[image] = comp_sum

print(vert_comp_star)
lowest_vals = [4]
old_lowest_vals = []
while cmp(lowest_vals, old_lowest_vals) != 0:
    
    #Calculate comparison 
    comp_mags = np.zeros(len(good_picsgood_pics))
    for image in good_picsgood_pics:
        comp_sum = 0
        for obj in lowest_vals:
            comp_sum += (all_SNR[image-1, obj]**2)*corrected_curves[obj][image]
        comp_mags[image] = comp_sum/norm_consts[image]
    
    #Calculate differnetial magnitude
    corrected_curves = -1(corrected_curves - comp_mags)
        
    #Calculate Weights WRONG
    norm_consts = np.zeros(len(good_picsgood_pics))
    for image in good_picsgood_pics:
        SNR2sum = 0
        for obj in lowest_vals: 
            SNR2sum += all_SNR[image-1,obj]**2
        norm_consts[image] = np.std(corrected_curves[obj])
        
    
    all_viability = []
    for source in sources:
        new = comp_mags-corrected_curves[source-1]
        all_viability.append(np.std(new))
    old_lowest_vals, lowest_vals = lowest_vals, list(np.argpartition(np.array(all_viability),5)[0:7])


# In[ ]:




# In[ ]:

#lowest_vals = [23, 35, 40, 25, 27, 29, 34]


# ##Current way of differential photometry is plain wrong.

# In[ ]:

get_ipython().magic(u'matplotlib inline')
star = 1

differential_curve = comp_mags-corrected_curves[source-1]
for source in sources:
    differential_curve = comp_mags-corrected_curves[source-1]
    """plt.plot(aij_raw['J.D.-2400000'], corrected_curves[source], 'bo')
    plt.title('Apass calibrated photometry')
    plt.xlabel('J.D.-2400000')
    plt.ylabel('Magnitude')
    #plt.gca().invert_yaxis()
    plt.savefig('Apass calibrated photometry')
    plt.show()"""
    plt.plot(aij_raw['J.D.-2400000'], differential_curve, 'bo')
    plt.title('Apass calibrated auto diff phot '+str(source))
    plt.xlabel('J.D.-2400000')
    plt.ylabel('Relative Magnitude')
    #plt.gca().invert_yaxis()
    plt.savefig('Apass calibrated automated differential photometry')
    plt.show()


# In[ ]:

table=Table([aij_raw['J.D.-2400000']+2400000, differential_curve, corrected_curves_er[1]], names=('JD','MAG','ERROR'))
print(table)


# In[ ]:

table.write('ETD_file.txt', format='ascii.tab')


# In[ ]:




# In[ ]:






from __future__ import print_function, division, absolute_import
import numpy as np
from photutils import aperture_photometry, CircularAperture, CircularAnnulus
from astropy.stats import sigma_clipped_stats
from astropy.nddata import NoOverlapError
from .coordinates import convert_pixel_wcs

__all__ = ['photutils_stellar_photometry',
           'detect_sources', 'faster_sigma_clip_stats',
           'find_too_close', 'clipped_sky_per_pix_stats',
           'add_to_photometry_table', 'photometry_on_directory']


def photutils_stellar_photometry(ccd_image, sources,
                                 aperture_radius, inner_annulus,
                                 outer_annulus, gain=1.0, N_R=0, N_dark_pp=0,
                                 reject_background_outliers=True):
    """
    Perform aperture photometry on an image, with a few options for estimating
    the local background from an annulus around the aperture.

    Parameters
    ----------
    ccd_image : `~ccdproc.CCDData`
        Image on which to perform aperture photometry.

    sources : `~astropy.table.Table`
        Table of extracted sources. Assumed to be the output of
        `~photutils.daofind()` source extraction function.

    aperture_radius : float
        Radius of aperture(s) in pixels.

    inner_annulus : float
        Inner radius of the annulus in pixels.

    outer_annulus : float
        Outer radius of the annulus in pixels.

    gain : float
        Gain of the CCD. In units of electrons per DN.

    N_R : float
        Read noise of the CCD in electrons per pixel.

    N_dark_pp : float
        Number of dark counts per pixel.

    reject_background_outliers : bool, optional
        If ``True``, sigma clip the pixels in the annulus to reject outlying
        pixels (e.g. like stars in the annulus)

    Returns
    -------
    phot_table : `~astropy.table.Table`
        Astropy table with columns for flux, x/y coordinates of center,
        RA/dec coordinates of center, sky background per pixel,
        net flux, aperture and annulus radii used, and flux error.
    """

    # check that the outer radius is greater or equal the inner radius
    # for annulus
    if inner_annulus >= outer_annulus:
        raise ValueError("outer_annulus must be greater than inner_annulus")

    # check that the annulus inner radius is greater or equal
    # the aperture radius
    if aperture_radius >= inner_annulus:
        raise ValueError("inner_radius must be greater than aperture_radius")

    # Extract x,y coordinates from sources table, construct aperture and
    # annulus objects from coordinates, and perform aperture photometry
    coords = (sources['xcentroid'], sources['ycentroid'])
    apertures = CircularAperture(coords, aperture_radius)
    annulus = CircularAnnulus(coords, inner_annulus, outer_annulus)
    phot_table = aperture_photometry(ccd_image, apertures)
    phot_table_1 = aperture_photometry(ccd_image, annulus)

    # Obtain the local background/pixel and net flux between the aperture and
    # annulus objects
    n_pix_ap = apertures.area()
    n_pix_ann = annulus.area()

    if reject_background_outliers:
        annulus_masks = annulus.to_mask()
        bkgd_pp = []
        for mask in annulus_masks:
            annulus_data = mask.cutout(ccd_image)
            bool_mask = mask.array < 1
            # Only include whole pixels in the estimate of the background
            bkgd, _, _ = sigma_clipped_stats(annulus_data * mask,
                                             mask=bool_mask)
            bkgd_pp.append(bkgd)
        bkgd_pp = np.array(bkgd_pp)
    else:
        bkgd_pp = phot_table_1['aperture_sum'] / n_pix_ann

    net_flux = phot_table['aperture_sum'] - (n_pix_ap * bkgd_pp)
    phot_table['background_per_pixel'] = bkgd_pp
    phot_table['net_flux'] = net_flux

    # Return a columns with the aperture radius and
    # the inner/outer annulus radii
    phot_table['aperture_radius'] = \
        np.ones(len(phot_table['aperture_sum'])) * aperture_radius
    phot_table['inner_radius'] = \
        np.ones(len(phot_table['aperture_sum'])) * inner_annulus
    phot_table['outer_radius'] = \
        np.ones(len(phot_table['aperture_sum'])) * outer_annulus

    # Obtain RA/Dec coordinates and add them to table
    try:
        ra, dec = convert_pixel_wcs(ccd_image, coords[0], coords[1], 1)
        phot_table['RA_center'] = ra
        phot_table['Dec_center'] = dec
    except AttributeError:
        pass

    # Obtain flux error and add column to return table
    noise = np.sqrt(gain * net_flux + n_pix_ap * (1 + (n_pix_ap / n_pix_ann)) *
                    (gain * (bkgd_pp + N_dark_pp) + (N_R**2)))
    phot_table['aperture_sum_err'] = noise

    return phot_table



# coding: utf-8

# In[1]:


import numpy as np
import bottleneck as bn

from astropy.coordinates import SkyCoord
from astropy.table import vstack
from astropy import units as u
from astropy.time import Time

from ccdproc import ImageFileCollection

from photutils import (DAOStarFinder, CircularAperture, CircularAnnulus,
                       aperture_photometry, centroid_sources)


np.seterr(all='ignore')

# In[11]:


def faster_sigma_clip_stats(data, sigma=5, iters=5, axis=None):

    data = data.copy()
    for _ in range(iters):
        central = bn.nanmedian(data, axis=axis)
        try:
            central = central[:, np.newaxis]
        except (ValueError, IndexError, TypeError):
            pass

        std_dif = 1.4826 * bn.nanmedian(np.abs(data - central))

        clips = np.abs(data - central) / std_dif > sigma

        if np.nansum(clips) == 0:
            break
        data[clips] = np.nan
    return (bn.nanmean(data, axis=axis), bn.nanmedian(data, axis=axis),
            bn.nanstd(data, axis=axis))


def detect_sources(ccd, fwhm, thresh):

    men, med, std = faster_sigma_clip_stats(ccd.data, sigma=5)


    # This sets up the source detection. The FWHM turns out to be key...making it too small results in a single star being detected as two separate sources.
    #
    # Stars must be brighter than the threshold to count as sources. Making the number higher gives you fewer detected sources, lower gives you more. There is no "magic" number.

    # In[13]:


    dao = DAOStarFinder(threshold=10 * std, fwhm=8, exclude_border=True)


    # Actually detect the stars...

    # In[14]:


    stars = dao(ccd - med)
    return stars


def find_too_close(star_locs, aperture_rad):
    star_coords = SkyCoord(ra=star_locs[0], dec=star_locs[1],
                           frame='icrs', unit='degree')
    idxc, d2d, d3d = star_coords.match_to_catalog_sky(star_coords,
                                                      nthneighbor=2)
    return d2d < (aperture_rad * 2 * 0.563 * u.arcsec)


def clipped_sky_per_pix_stats(data, annulus, sigma=5, iters=5):
    # Get a list of masks from the annuli
    # Use the 'center' method because then pixels are either in or out. To use
    # 'partial' or 'exact' we would need to do a weighted sigma clip and
    # I'm not sure how to do that.
    masks = annulus.to_mask(method='center')

    anul_list = []
    for mask in masks:
        # Multiply the mask times the data
        to_clip = mask.multiply(data.data, fill_value=np.nan)
        anul_list.append(to_clip.flatten())
    # Convert the list to an array for doing the sigma clip
    anul_array = np.array(anul_list)
    # Turn all zeros into np.nan...
    anul_array[anul_array == 0] = np.nan
    avg_sky_per_pix, med_sky_per_pix, std_sky_per_pix = \
        faster_sigma_clip_stats(anul_array,
                                sigma=sigma,
                                iters=iters,
                                axis=1
                               )

    return (avg_sky_per_pix * data.unit, med_sky_per_pix * data.unit,
            std_sky_per_pix * data.unit)


# ### Add more columns to the data table

# In[27]:


def add_to_photometry_table(phot, ccd, annulus, apertures, fname='',
                            star_ids=None, gain=None):
    phot.rename_column('aperture_sum_0', 'aperture_sum')
    phot['aperture_sum'].unit = u.adu
    phot.rename_column('aperture_sum_1', 'annulus_sum')
    star_locs = ccd.wcs.all_pix2world(phot['xcenter'], phot['ycenter'], 0)
    star_coords = SkyCoord(ra=star_locs[0], dec=star_locs[1],
                           frame='icrs', unit='degree')
    phot['RA'] = star_coords.ra
    phot['Dec'] = star_coords.dec
    print('        ...calculating clipped sky stats')
    avg_sky_per_pix, med_sky_per_pix, std_sky_per_pix = \
        clipped_sky_per_pix_stats(ccd, annulus)
    print('        ...DONE calculating clipp sky stats')
    phot['sky_per_pix_avg'] = avg_sky_per_pix
    phot['sky_per_pix_med'] = med_sky_per_pix
    phot['sky_per_pix_std'] = std_sky_per_pix
    phot['aperture'] = apertures.r * u.pixel
    phot['aperture_area'] = apertures.area()  # * u.pixel * u.pixel
    phot['annulus_inner'] = annulus.r_in * u.pixel
    phot['annulus_outer'] = annulus.r_out * u.pixel
    phot['annulus_area'] = annulus.area()  # * u.pixel * u.pixel
    phot['exposure'] = [ccd.header['exposure']] * len(phot) * u.second
    phot['date-obs'] = [ccd.header['DATE-OBS']] * len(phot)
    night = Time(ccd.header['DATE-OBS'], scale='utc')
    night.format = 'mjd'
    phot['night'] = np.int(np.floor(night.value - 0.5))
    phot['aperture_net_flux'] = (phot['aperture_sum'] -
                                 (phot['aperture_area'] *
                                  phot['sky_per_pix_avg']))

    if gain is not None:
        phot['mag_inst_{}'.format(ccd.header['filter'])] = \
            (-2.5 * np.log10(gain * phot['aperture_net_flux'].value /
                             phot['exposure'].value))

    metadata_to_add = ['AIRMASS', 'FILTER']
    for meta in metadata_to_add:
        phot[meta.lower()] = [ccd.header[meta]] * len(phot)
    if fname:
        phot['file'] = fname
    if star_ids is not None:
        phot['star_id'] = star_ids


def photometry_on_directory(directory_with_images, object_of_interest,
                            star_locs, aperture_rad,
                            inner_annulus, outer_annulus,
                            max_adu, star_ids,
                            gain, read_noise, dark_current):
    ifc = ImageFileCollection(directory_with_images)
    phots = []
    missing_stars = []
    for a_ccd, fname in ifc.ccds(object=object_of_interest, return_fname=True):
        print('on image ', fname)
        try:
            # Convert RA/Dec to pixel coordinates for this image
            pix_coords = a_ccd.wcs.all_world2pix(star_locs[0], star_locs[1], 0)
        except AttributeError:
            print('    ....SKIPPING THIS IMAGE, NO WCS')
            continue

        xs, ys = pix_coords

        # Remove anything that is too close to the edges/out of frame
        padding = 3 * aperture_rad
        out_of_bounds = ((xs < padding) | (xs > (a_ccd.shape[1] - padding)) |
                         (ys < padding) | (ys > (a_ccd.shape[0] - padding)))
        in_bounds = ~out_of_bounds

        # Find centroids of each region around star that is in_bounds
        xs_in = xs[in_bounds]
        ys_in = ys[in_bounds]
        print('    ...finding centroids')
        try:
            xcen, ycen = centroid_sources(a_ccd.data, xs_in, ys_in,
                                          box_size=2 * aperture_rad + 1)
        except NoOverlapError:
            print('    ....SKIPPING THIS IMAGE, CENTROID FAILED')
            continue

        # Calculate offset between centroid in this image and the positions
        # based on input RA/Dec. Later we will set the magnitude of those with
        # large differences to an invalid value (maybe).
        center_diff = np.sqrt((xs_in - xcen)**2 + (ys_in - ycen)**2)

        # Set up apertures and annuli based on the centroids in this image.
        aps = CircularAperture((xcen, ycen), r=aperture_rad)
        anuls = CircularAnnulus((xcen, ycen), inner_annulus, outer_annulus)

        # Set any clearly bad values to NaN
        a_ccd.data[a_ccd.data > max_adu] = np.nan
        print('    ...doing photometry')
        # Do the photometry...
        pho = aperture_photometry(a_ccd.data, (aps, anuls),
                                  mask=a_ccd.mask, method='center')

        # We may have some stars we did not do photometry for because
        # those stars were out of bounds.
        # Add the ones we missed to the list of missing
        missed = star_ids[out_of_bounds]
        missing_stars.append(missed)

        # Add all the extra goodies to the table
        print('    ...adding extra columns')
        add_to_photometry_table(pho, a_ccd, anuls, aps,
                                fname=fname, star_ids=star_ids[in_bounds],
                                gain=1.47)
        # And add the final table to the list of tables
        phots.append(pho)

    # ### Combine all of the individual photometry tables into one

    all_phot = vstack(phots)

    # ### Eliminate any stars that are missing from one or more images
    #
    # This makes life a little easier later...

    uniques = set()
    for miss in missing_stars:
        uniques.update(set(miss))

    actually_bad = sorted([u for u in uniques if u in all_phot['star_id']])
    len(uniques), len(actually_bad)

    all_phot.add_index('star_id')
    if actually_bad:
        bad_rows = all_phot.loc_indices[actually_bad]
        try:
            bad_rows = list(bad_rows)
        except TypeError:
            bad_rows = [bad_rows]
        all_phot.remove_indices('star_id')
        all_phot.remove_rows(sorted(bad_rows))

    all_phot.remove_indices('star_id')

    snr = (gain * all_phot['aperture_net_flux'] /
           np.sqrt(gain * all_phot['aperture_net_flux'].value +
                   all_phot['aperture_area'] *
                   (1 + all_phot['aperture_area'] / all_phot['annulus_area']) *
                   (gain * all_phot['sky_per_pix_avg'].value +
                    gain * dark_current * all_phot['exposure'].value +
                    read_noise**2
                   )
                  ))

    all_phot['mag_error'] = 1.085736205 / snr

    return all_phot
